## <>
module SipHash
	class << self
		## int64, int64, int64, int64 -> [int64, int64, int64, int64]
		def __round(v0, v1, v2, v3)
			v0 += v1
			v2 += v3
			v1 = v1.rol(13)
			v3 = v3.rol(16)
			v1 ^= v0
			v3 ^= v2
			v0 = v0.rol(32)
			v2 += v1
			v0 += v3
			v1 = v1.rol(17)
			v3 = v3.rol(21)
			v1 ^= v2
			v3 ^= v0
			v2 = v2.rol(32)
			return [v0, v1, v2, v3]
		end

		## int64, int64, byte[] -> int64
		def digest_as_int64(k0, k1, m)
			v0 = k0 ^ 0x736F6D6570736575
			v1 = k1 ^ 0x646F72616E646F6D
			v2 = k0 ^ 0x6C7967656E657261
			v3 = k1 ^ 0x7465646279746573
			n = m.length.i32 & ~7.i32
			i = 0.i32
			while i < n
				w = 0.i64
				j = 0.i32
				while j < 64
					w |= (m[i.as_i].byte & 0xFF).to_int64 << j
					i += 1
					j += 8
				end
				v3 ^= w
				v0, v1, v2, v3 = __round(v0, v1, v2, v3)
				v0, v1, v2, v3 = __round(v0, v1, v2, v3)
				v0 ^= w
			end
			w = 0.i64
			j = 0.i32
			while i < m.length
				w |= (m[i.as_i].byte & 0xFF).to_int64 << j
				i += 1
				j += 8
			end
			w |= m.length.i32.to_int64 << 070
			v3 ^= w
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			v0 ^= w
			v2 ^= 0xFF
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			v0, v1, v2, v3 = __round(v0, v1, v2, v3)
			return v0 ^ v1 ^ v2 ^ v3
		end

		## String, String -> Integer
		def digest(k, s)
			m = k.to_java_bytes_unsafe
			n = m.length.i32
			if n != 020
				raise ArgumentError, "key must be 16 bytes"
			end
			k1 = 0.i64
			while n > 8
				n -= 1
				k1 <<= 8
				k1 |= (m[n.as_i].byte & 0xFF).to_int64
			end
			k0 = 0.i64
			while n > 0
				n -= 1
				k0 <<= 8
				k0 |= (m[n.as_i].byte & 0xFF).to_int64
			end
			v = digest_as_int64(k0, k1, s.to_java_bytes_unsafe)
			if v < 0
				return v.new_fixnum & 0xFFFFFFFFFFFFFFFF
			end
			return v.new_fixnum
		end
	end
end
